/**
 * @file
 * @author RECORD Team,
 *   http://record.toulouse.inra.fr ,
 *   Applied Mathematics and Informatics Research Division &
 *   Environment and Agronomy Research Division,
 *   French National Institute for Agricultural Research
 * @version $Id: $
 */

/*
 * Copyright (C) 2011 - INRA
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#define BOOST_TEST_MAIN
#define BOOST_AUTO_TEST_MAIN
#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE test_toto

#include <boost/test/unit_test.hpp>
#include <boost/test/auto_unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/assert.hpp>

#include <fstream>
#include <iterator>
#include <iostream>
#include <iomanip>      // std::setprecision

#include <vle/devs/Dynamics.hpp>
#include <vle/utils/Package.hpp>
#include <vle/vpz/Vpz.hpp>
#include <boost/numeric/conversion/cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/predicate.hpp>

namespace vd = vle::devs;
namespace vv = vle::value;
namespace vz = vle::vpz;
namespace vu = vle::utils;

namespace Tester {
/**
 * @brief a Difference Equation test model based on a Dynamics
 *
 * A simple model enabling to test Difference Equation.
 * A test is first build by running the model in the mode storage. At
 * this time all the events storing Variables of the Difference
 * Equation are stored. Then if you change the mode to check, when
 * running the model, each variable received is compared against the
 * file.
 *
 * To be run in the check mode, a specific "tester" command is
 * provided by the tester package
 *
 * the list of ports :
 * - Experimental condition :
 *  - file : [String][mandatory] a file where to store the data for the test
 *  - action [String][mandatory] the action to be done by the model
 *   - store : storing the data of the test
 *   - check : checking the simulation according to the file
 *  - tolerance : the tolerance we accept in the value comparison
 * - Input :
 *  - "Free" : port names are not used, the user can do what ever he
 * want, but the events should only contain Variables Values of
 * Difference Equations.
 * - Output :
 * - Observation :
 */
class DETester: public vd::Dynamics
{

public:
    DETester(const vd::DynamicsInit& mdl,
                     const vd::InitEventList& evts)
        : vd::Dynamics(mdl, evts)
    {
        dataPackage = vv::toString(evts.get("PkgName"));
        vu::Package pack(dataPackage);
        dataTestFile = pack.getDataFile(vv::toString(evts.get("file")));

        action = vv::toString(evts.get("action"));

        if (action == "store") {
            testOutFile.open (dataTestFile.c_str(), std::ios::out);
        } else if (action == "check") {
            testOutFile.open (dataTestFile.c_str(), std::ios::in);
        }

        if (evts.exist("tolerance")) {
            tolerance = vv::toDouble(evts.get("tolerance"));
        } else {
            tolerance = 0.00000000001;
        }

    };

    virtual ~DETester() {
         testOutFile.close();
    };

    virtual vd::Time init(const vd::Time& /*time*/)
    {
        return vd::infinity;
    };

    vd::Time timeAdvance() const
    {
        return vd::infinity;
    };

    /**
     * @brief either store or compare values and variables
     */
    void externalTransition(
        const vd::ExternalEventList& events,
        const vd::Time& time)
    {
        for (vd::ExternalEventList::const_iterator it = events.begin();
             it != events.end(); ++it) {
            double value  = (*it)->getDoubleAttributeValue("value");
            std::string name = (*it)->getStringAttributeValue("name");
            if (action == "store") {

                testOutFile << boost::str(boost::format("%1%;%2%;%3%\n") %
                                          boost::io::group(std::setprecision(std::numeric_limits<double>::digits10), time) %
                                          name %
                                          boost::io::group(std::setprecision(std::numeric_limits<double>::digits10), value));
            } else if (action == "check") {
                std::string line;
                testOutFile >> line;

                std::vector< std::string > SplitVec;
                boost::split( SplitVec, line , boost::is_any_of(";"));

                BOOST_REQUIRE_EQUAL((double)time, boost::lexical_cast<double>(SplitVec[0]));
                BOOST_REQUIRE_EQUAL(name, SplitVec[1]);
                BOOST_REQUIRE_CLOSE(value, boost::lexical_cast<double>(SplitVec[2]), tolerance);
            }
        };
    };

private:

    std::fstream testOutFile;

    std::string action;
    std::string dataTestFile;
    std::string dataPackage;

    double tolerance;
};

} // namespace Tester

DECLARE_DYNAMICS(Tester::DETester)
