/**
 * @file tester/TesterSimulation.hpp
 * @author The VLE Development Team
 */

/*
 * VLE Environment - the multimodeling and simulation environment
 * This file is a part of the VLE environment (http://vle.univ-littoral.fr)
 * Copyright (C) 2003 - 2009 The VLE Development Team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef TESTER_SIMULATION_HPP
#define TESTER_SIMULATION_HPP 1

#include <vle/vle.hpp>
#include <vle/value/Map.hpp>
#include <vle/value/Matrix.hpp>
#include <vle/value/Value.hpp>
#include <vle/vpz/Vpz.hpp>
#include <vle/vpz/Model.hpp>
#include <vle/vpz/AtomicModel.hpp>
#include <vle/utils/Package.hpp>
#include <vle/utils/Exception.hpp>
#include <vle/utils/Package.hpp>
#include <vle/manager/Simulation.hpp>
#include <vle/manager/Manager.hpp>

namespace tester {

namespace vz = vle::vpz;
namespace vu = vle::utils;
namespace vm = vle::manager;
namespace va = vle::value;

/**
 * @brief Provides high level functions for simulating a vpz from the current
 * package
 */
class TesterSimulation
{
public:
    /**
     * TesterSimulation constructor
     * @brief packagename, the name of the current package to test
     */
    TesterSimulation(const std::string& packagename):
        mapp(), mpackage(packagename), mvpz(0), msimres(0), mpack(packagename)
    {
    }

    ~TesterSimulation()
    {
        delete mvpz;
    }

    /**
     * @brief Get an element of a view
     * @param view, expected to be a view, ie, frst column gives the names
     * of the columns
     * @brief atomModel, the complete name of the atomic model for which
     * outputs are required
     * @brief portName, the observation port name
     * @brief i, the index of the element to get
     * @return the ith element of the column "atomModel.portName" of
     * the view
     */
    const va::Value& getViewElt(const va::Matrix& view,
            const std::string& atomModel, const std::string& portName,
            int i)
    {
        int col = -1;
        std::string path = atomModel;
        path.append(".");
        path.append(portName);

        for(unsigned int c=0; c<view.columns(); c++){
            if(view.getString(c,0) == path){
                col = c;
            }
        }
        return *view.get(col,i);
    }

    /**
     * @brief Set the current model to test
     */
    void setVpz(const std::string& vpzname)
    {
        delete mvpz;
        mvpz = new vz::Vpz(mpack.getExpFile(vpzname));
    }

    /**
     * @brief Attaches to an atomic model a set of conditions
     * @param atomModel, the complete name of the model
     * @param conds, the set of conditions to attach to atomModel
     */
    void setConditions(const std::string& atomModel,
            const std::vector<std::string>& conds)
    {
        if(mvpz){
            vz::Model& vpz_mod = mvpz->project().model();
            vz::BaseModel* mdl = vpz_mod.model()->findModelFromPath(atomModel);
            vz::AtomicModel* atomg = mdl->toAtomic();
            atomg->setConditions(conds);
        }
    }
    /**
     *@brief Set all views to storage mode in order to get results.
     *TODO dependence on vle.output
     */
    void setStorageViews()
    {
        if(mvpz){
            //set all output plugin to storage
            vz::Outputs::iterator itb =
                    mvpz->project().experiment().views().outputs().begin();
            vz::Outputs::iterator ite =
                    mvpz->project().experiment().views().outputs().end();
            for(;itb!=ite;itb++) {
                va::Map* configOutput = new va::Map();
                //configOutput->addInt("rows",10000);
                //configOutput->addInt("inc_rows",10000);
                configOutput->addString("header","top");//A voir
                vz::Output& output = itb->second;
                output.setLocalStream("", "storage", "vle.output");
                output.setData(configOutput);
            }
        }
    }
    /**
     * @brief Get all views for one simple simulation
     * @brief vpzname, the name of the vpz of the current package
     * @return the map corresponding
     */
    const va::Map& simulates(const std::string& vpzname)
    {
        delete msimres;
        setVpz(vpzname);
        setStorageViews();

        //simulation
        vu::ModuleManager man;
        vm::Error error;
        vm::Simulation sim(vm::LOG_NONE, vm::SIMULATION_NONE, NULL);
        msimres = sim.run(new vz::Vpz(*mvpz), man, &error);
        return *msimres;
    }

    /**
     * @brief Get one view for one simple simulation
     * @brief vpzname, the name of the vpz of the current package
     * @brief viewname, the view to return
     * @return the matrix corresponding
     */
    const va::Matrix& simulates(const std::string& vpzname,
            const std::string& viewname)
    {
        const va::Map& views = simulates(vpzname);
        return views.getMatrix(viewname);
    }

    /**
     * @brief One simple simulation
     * @brief vpzname, the name of the vpz of the current package
     * @brief viewname, the view to return
     * @brief atomModel, the complete name of the atomic model for which
     * outputs are required
     * @brief portName, the observation port name
     * @return the vector corresponding
     */
    va::ConstVectorView simulates(const std::string& vpzname,
            const std::string& viewname, const std::string& atomModel,
            const std::string& portName)
    {
        simulates(vpzname, viewname);

        const va::Matrix& view = msimres->getMatrix(viewname);

        std::string portToFind = (atomModel + std::string(".") + portName);

        for(unsigned int j=1; j < view.columns(); j++){
            if(view.getString(j,0).compare(portToFind) == 0){
                return view.column(j);
            }
        }
        throw vu::ArgError( vle::fmt(" Model.Port not found : %1%") % portToFind);
        return view.column(0);

    }

    /**
     * @brief One simple simulation
     * @brief vpzname, the name of the vpz of the current package
     * @brief viewname, the view to return
     * @brief atomModel, the complete name of the atomic model for which
     * outputs are required
     * @brief portName, the observation port name
     * @brief index, the index of the vlaue to get in column
     * resulting from the selection of view * atomModel * portName
     * @return the value corresponding
     */
    const va::Value* simulates(const std::string& vpzname,
            const std::string& viewname, const std::string& atomModel,
            const std::string& portName, unsigned int index)
    {
        const va::ConstVectorView col = simulates(vpzname, viewname,
                atomModel, portName);
        if(col[index]){
            return col[index];
        } else {
            return NULL;
        }
    }

private:
    vle::Init mapp;
    std::string mpackage;
    vz::Vpz* mvpz;
    va::Map* msimres;
    vu::Package mpack;

};

}//namespace
#endif
